@@ -285,6 +285,12 @@ def node_power_action(task, new_state, timeout=None):
         driver_internal_info['last_power_state_change'] = str(
             timeutils.utcnow().isoformat())
         node.driver_internal_info = driver_internal_info
+        # NOTE(dtantsur): wipe token on shutting down, otherwise a reboot in
+        # fast-track (or an accidentally booted agent) will cause subsequent
+        # actions to fail.
+        if target_state in (states.POWER_OFF, states.SOFT_POWER_OFF,
+                            states.REBOOT, states.SOFT_REBOOT):
+            wipe_internal_info_on_power_off(node)
         node.save()
 
     # take power action
@@ -444,6 +450,19 @@ def cleaning_error_handler(task, msg, tear_down_cleaning=True,
         task.process_event('fail', target_state=target_state)
 
 
+def wipe_internal_info_on_power_off(node):
+    """Wipe information that should not survive reboot/power off."""
+    driver_internal_info = node.driver_internal_info
+    if not is_agent_token_pregenerated(node):
+        # Wipe the token if it's not pre-generated, otherwise we'll refuse to
+        # generate it again for the newly booted agent.
+        driver_internal_info.pop('agent_secret_token', False)
+    # Wipe cached steps since they may change after reboot.
+    driver_internal_info.pop('agent_cached_deploy_steps', None)
+    driver_internal_info.pop('agent_cached_clean_steps', None)
+    node.driver_internal_info = driver_internal_info
+
+
 def wipe_token_and_url(task):
     """Remove agent URL and token from the task."""
     info = task.node.driver_internal_info
@@ -961,6 +980,21 @@ def value_within_timeout(value, timeout):
     return last_valid <= last
 
 
+def agent_is_alive(node, timeout=None):
+    """Check that the agent is likely alive.
+
+    The method then checks for the last agent heartbeat, and if it occured
+    within the timeout set by [deploy]fast_track_timeout, then agent is
+    presumed alive.
+
+    :param node: A node object.
+    :param timeout: Heartbeat timeout, defaults to `fast_track_timeout`.
+    """
+    return value_within_timeout(
+        node.driver_internal_info.get('agent_last_heartbeat'),
+        timeout or CONF.deploy.fast_track_timeout)
+
+
 def is_fast_track(task):
     """Checks a fast track is available.
 
@@ -985,9 +1019,7 @@ def is_fast_track(task):
                       {'node': task.node.uuid, 'error': task.node.last_error})
         return False
 
-    if value_within_timeout(
-            task.node.driver_internal_info.get('agent_last_heartbeat'),
-            CONF.deploy.fast_track_timeout):
+    if agent_is_alive(task.node):
         return True
     else:
         LOG.debug('Node %(node)s should be fast-track-able, but the agent '
@@ -1194,4 +1226,7 @@ def get_attached_vif(port):
     rescue_vif = port.internal_info.get('rescuing_vif_port_id')
     if rescue_vif:
         return (rescue_vif, 'rescuing')
+    inspection_vif = port.internal_info.get('inspection_vif_port_id')
+    if inspection_vif:
+        return (inspection_vif, 'inspecting')
     return (None, None)
