@@ -35,6 +35,7 @@ from ironic.conductor import steps as conductor_steps
 from ironic.conductor import task_manager
 from ironic.conductor import utils as manager_utils
 from ironic.conf import CONF
+from ironic.drivers import base
 from ironic.drivers.modules import agent_client
 from ironic.drivers.modules import boot_mode_utils
 from ironic.drivers.modules import deploy_utils
@@ -71,7 +72,14 @@ VENDOR_PROPERTIES = {
     'deploy_forces_oob_reboot': _(
         'Whether Ironic should force a reboot of the Node via the out-of-band '
         'channel after deployment is complete. Provides compatibility with '
-        'older deploy ramdisks. Defaults to False. Optional.')
+        'older deploy ramdisks. Defaults to False. Optional.'),
+    'agent_verify_ca': _(
+        'Either a Boolean value, a path to a CA_BUNDLE file or directory with '
+        'certificates of trusted CAs. If set to True ironic will verify '
+        'the agent\'s certificate; if False the driver will ignore verifying '
+        'the SSL certificate. If it\'s a path the driver will use the '
+        'specified certificate or one of the certificates in the '
+        'directory. Defaults to True. Optional'),
 }
 
 __HEARTBEAT_RECORD_ONLY = (states.ENROLL, states.MANAGEABLE, states.AVAILABLE,
@@ -335,6 +343,13 @@ def get_steps(task, step_type, interface=None, override_priorities=None):
     return steps
 
 
+def find_step(task, step_type, interface, name):
+    """Find the given in-band step."""
+    steps = get_steps(task, step_type, interface)
+    return conductor_steps.find_step(
+        steps, {'interface': interface, 'step': name})
+
+
 def _raise(step_type, msg):
     assert step_type in ('clean', 'deploy')
     exc = (exception.NodeCleaningFailure if step_type == 'clean'
@@ -342,18 +357,20 @@ def _raise(step_type, msg):
     raise exc(msg)
 
 
-def execute_step(task, step, step_type):
+def execute_step(task, step, step_type, client=None):
     """Execute a clean or deploy step asynchronously on the agent.
 
     :param task: a TaskManager object containing the node
     :param step: a step dictionary to execute
     :param step_type: 'clean' or 'deploy'
+    :param client: agent client (if available)
     :raises: NodeCleaningFailure (clean step) or InstanceDeployFailure (deploy
         step) if the agent does not return a command status.
     :returns: states.CLEANWAIT/DEPLOYWAIT to signify the step will be
         completed async
     """
-    client = _get_client()
+    if client is None:
+        client = _get_client()
     ports = objects.Port.list_by_node_id(
         task.context, task.node.id)
     call = getattr(client, 'execute_%s_step' % step_type)
@@ -382,8 +399,21 @@ def _step_failure_handler(task, msg, step_type):
 class HeartbeatMixin(object):
     """Mixin class implementing heartbeat processing."""
 
+    has_decomposed_deploy_steps = False
+    """Whether the driver supports decomposed deploy steps.
+
+    Previously (since Rocky), drivers used a single 'deploy' deploy step on
+    the deploy interface. Some additional steps were added for the 'direct'
+    and 'iscsi' deploy interfaces in the Ussuri cycle, which means that
+    more of the deployment flow is driven by deploy steps.
+    """
+
     def __init__(self):
         self._client = _get_client()
+        if not self.has_decomposed_deploy_steps:
+            LOG.warning('%s does not support decomposed deploy steps. This '
+                        'is deprecated and will stop working in a future '
+                        'release', self.__class__.__name__)
 
     def continue_deploy(self, task):
         """Continues the deployment of baremetal node.
@@ -501,8 +531,12 @@ class HeartbeatMixin(object):
             # are currently in the core deploy.deploy step. Other deploy steps
             # may cause the agent to boot, but we should not trigger deployment
             # at that point if the driver is polling for completion of a step.
-            if self.in_core_deploy_step(task):
+            if (not self.has_decomposed_deploy_steps
+                    and self.in_core_deploy_step(task)):
                 msg = _('Failed checking if deploy is done')
+                # NOTE(mgoddard): support backwards compatibility for
+                # drivers which do not implement continue_deploy and
+                # reboot_to_instance as deploy steps.
                 if not self.deploy_has_started(task):
                     msg = _('Node failed to deploy')
                     self.continue_deploy(task)
@@ -740,7 +774,60 @@ class AgentBaseMixin(object):
             task, manage_boot=self.should_manage_boot(task))
 
 
-class AgentDeployMixin(HeartbeatMixin):
+class AgentOobStepsMixin(object):
+    """Mixin with out-of-band deploy steps."""
+
+    @METRICS.timer('AgentDeployMixin.switch_to_tenant_network')
+    @base.deploy_step(priority=30)
+    @task_manager.require_exclusive_lock
+    def switch_to_tenant_network(self, task):
+        """Deploy step to switch the node to the tenant network.
+
+        :param task: a TaskManager object containing the node
+        """
+        try:
+            with manager_utils.power_state_for_network_configuration(task):
+                task.driver.network.remove_provisioning_network(task)
+                task.driver.network.configure_tenant_networks(task)
+        except Exception as e:
+            msg = (_('Error changing node %(node)s to tenant networks after '
+                     'deploy. %(cls)s: %(error)s') %
+                   {'node': task.node.uuid, 'cls': e.__class__.__name__,
+                    'error': e})
+            # NOTE(mgoddard): Don't collect logs since the node has been
+            # powered off.
+            log_and_raise_deployment_error(task, msg, collect_logs=False,
+                                           exc=e)
+
+    @METRICS.timer('AgentDeployMixin.boot_instance')
+    @base.deploy_step(priority=20)
+    @task_manager.require_exclusive_lock
+    def boot_instance(self, task):
+        """Deploy step to boot the final instance.
+
+        :param task: a TaskManager object containing the node
+        """
+        can_power_on = (states.POWER_ON in
+                        task.driver.power.get_supported_power_states(task))
+        try:
+            if can_power_on:
+                manager_utils.node_power_action(task, states.POWER_ON)
+            else:
+                LOG.debug('Not trying to power on node %s that does not '
+                          'support powering on, assuming already running',
+                          task.node.uuid)
+        except Exception as e:
+            msg = (_('Error booting node %(node)s after deploy. '
+                     '%(cls)s: %(error)s') %
+                   {'node': task.node.uuid, 'cls': e.__class__.__name__,
+                    'error': e})
+            # NOTE(mgoddard): Don't collect logs since the node has been
+            # powered off.
+            log_and_raise_deployment_error(task, msg, collect_logs=False,
+                                           exc=e)
+
+
+class AgentDeployMixin(HeartbeatMixin, AgentOobStepsMixin):
     """Mixin with deploy methods."""
 
     @METRICS.timer('AgentDeployMixin.get_clean_steps')
@@ -1067,16 +1154,13 @@ class AgentDeployMixin(HeartbeatMixin):
             LOG.error(msg)
             return _step_failure_handler(task, msg, step_type)
 
-    @METRICS.timer('AgentDeployMixin.reboot_and_finish_deploy')
-    def reboot_and_finish_deploy(self, task):
-        """Helper method to trigger reboot on the node and finish deploy.
-
-        This method initiates a reboot on the node. On success, it
-        marks the deploy as complete. On failure, it logs the error
-        and marks deploy as failure.
+    @METRICS.timer('AgentDeployMixin.tear_down_agent')
+    @base.deploy_step(priority=40)
+    @task_manager.require_exclusive_lock
+    def tear_down_agent(self, task):
+        """A deploy step to tear down the agent.
 
         :param task: a TaskManager object containing the node
-        :raises: InstanceDeployFailure, if node reboot failed.
         """
         wait = CONF.agent.post_deploy_get_power_state_retry_interval * 1000
         attempts = CONF.agent.post_deploy_get_power_state_retries + 1
@@ -1098,9 +1182,17 @@ class AgentDeployMixin(HeartbeatMixin):
         # in-band methods
         oob_power_off = strutils.bool_from_string(
             node.driver_info.get('deploy_forces_oob_reboot', False))
+        can_power_on = (states.POWER_ON in
+                        task.driver.power.get_supported_power_states(task))
 
         try:
-            if not oob_power_off:
+            if not can_power_on:
+                LOG.info('Power interface of node %(node)s does not support '
+                         'power on, using reboot to switch to the instance',
+                         node.uuid)
+                self._client.sync(node)
+                manager_utils.node_power_action(task, states.REBOOT)
+            elif not oob_power_off:
                 try:
                     self._client.power_off(node)
                 except Exception as e:
@@ -1145,23 +1237,21 @@ class AgentDeployMixin(HeartbeatMixin):
                     'error': e})
             log_and_raise_deployment_error(task, msg, exc=e)
 
-        try:
-            with manager_utils.power_state_for_network_configuration(task):
-                task.driver.network.remove_provisioning_network(task)
-                task.driver.network.configure_tenant_networks(task)
-            manager_utils.node_power_action(task, states.POWER_ON)
-        except Exception as e:
-            msg = (_('Error rebooting node %(node)s after deploy. '
-                     '%(cls)s: %(error)s') %
-                   {'node': node.uuid, 'cls': e.__class__.__name__,
-                    'error': e})
-            # NOTE(mgoddard): Don't collect logs since the node has been
-            # powered off.
-            log_and_raise_deployment_error(task, msg, collect_logs=False,
-                                           exc=e)
+    # TODO(dtantsur): remove in W
+    @METRICS.timer('AgentDeployMixin.reboot_and_finish_deploy')
+    def reboot_and_finish_deploy(self, task):
+        """Helper method to trigger reboot on the node and finish deploy.
 
-        # TODO(dtantsur): remove these two calls when this function becomes a
-        # real deploy step.
+        This method initiates a reboot on the node. On success, it
+        marks the deploy as complete. On failure, it logs the error
+        and marks deploy as failure.
+
+        :param task: a TaskManager object containing the node
+        :raises: InstanceDeployFailure, if node reboot failed.
+        """
+        # NOTE(dtantsur): do nothing here, the new deploy steps tear_down_agent
+        # and boot_instance will be picked up and finish the deploy (even for
+        # legacy deploy interfaces without decomposed steps).
         task.process_event('wait')
         manager_utils.notify_conductor_resume_deploy(task)
 
